#!/usr/bin/env python2
# coding: utf-8
from __future__ import print_function, unicode_literals
import traceback
import struct
import time
import sys
import re
import os

"""vdi2raw.py: stream vm hdd to bare metal"""
__version__   = "1.1"
__author__    = "ed <irc.rizon.net>"
__url__       = "https://ocv.me/dev/?vdi2raw.py"
__credits__   = ["stackoverflow.com"]
__license__   = "MIT"
__copyright__ = 2017

"""
this supports both py2 and py3
but as usual py2 is faster
"""

VERBOSE = False

def p(*args, **kwargs):
	print(*args, file=sys.stderr, **kwargs)

def dbg(*args):
	stack = traceback.extract_stack()
	filename, lineno, function_name, code = stack[-2]
	varnames = re.compile(r'^[^(]*\((.*?)\)[^)]*$').search(code).groups()[0]
	varnames = varnames.split(',')
	for arg in args:
		varname = varnames.pop(0).strip()
		p('{0:12} {1:x}'.format(varname[:12], arg))

def uuid(h):
	h = struct.unpack(b'<LHH',h[:8]) + struct.unpack(b'>HHL',h[8:])
	return '{:08x}-{:04x}-{:04x}-{:04x}-{:04x}{:08x}'.format(h[0],h[1],h[2],h[3],h[4],h[5])

def human(v):
	suffix = 'byte'
	for unit in ['KB','MB','GB','TB','PB','EB','ZB']:
		if abs(v) >= 1024:
			v /= 1024.0
			suffix = unit
	width = 0
	if v < 1000: width = 1
	if v < 100: width = 2
	if v < 10: width = 3
	return '{v:.{w}f} {s}'.format(v=v, w=width, s=suffix)

p()
if len(sys.argv) < 2:
	self = os.path.split(sys.argv[0])[-1]
	p('usage 1: {0} image.vdi'.format(self))
	p('  writes entire disk image to STDOUT,')
	p('  also supports the following if your python does:')
	p('     vdi.gz  vdi.bz2  vdi.xz  vdi.lzma')
	p()
	p('usage 2: {0} image.vdi /dev/sda'.format(self))
	p('  writes sparse disk image to physical device')
	p('  (skips each 1MB block which is all-zero)')
	p()
	p('this goes well with revert.py:')
	p('  {0} image.vdi.gz | revert.py /dev/sda'.format(self))
	p()
	exit(0)

stdout = sys.stdout
if sys.version_info >= (3,0):
	stdout = stdout.buffer

is_stdout = True
target = stdout
if len(sys.argv) >= 3:
	is_stdout = False
	target = open(sys.argv[2], 'wb')
	p('writing to file or device using sparse mode,')
	p('entire 1MB blocks of zeroed bytes will be skipped.')
	p('Tiny chance of corruption when writing to a device.')
	p()
else:
	p('writing to STDOUT, including all-zero blocks')

fn = sys.argv[1]
if fn.endswith('.vdi'):
	opener = open
elif fn.endswith('.gz'):
	import gzip
	opener = gzip.open
elif fn.endswith('.bz2'):
	import bz2
	opener = bz2.BZ2File
elif fn.endswith('.xz') or fn.endswith('.lzma'):
	import lzma
	opener = lzma.open
else:
	p('unsupported file extension [{0}]'.format(fn.split('.')[-1]))
	sys.exit(1)

vdi = opener(fn, 'rb')

ptr = 64
ptr2 = int(64 + 5*32/8 + 256 + 11*32/8 + 1*64/8 + 32*4 + 64*7/8)
header = vdi.read(ptr2)

part = struct.unpack(b'<LLLLL', header[ptr:ptr+20])
ptr += 20

h_magic = part[0]
h_version = part[1]
h_len_header = part[2]
h_img_type = part[3]
h_img_flags = part[4]

if VERBOSE:
	dbg(h_magic, h_version, h_len_header, h_img_type, h_img_flags)

if h_magic != 0xbeda107f:
	p('not a vdi file')
	exit()

if h_version != 0x00010001:
	p('unsupported vdi version')
	exit()

# skip description
ptr += 256

part = struct.unpack(b'<LLLLLLLQLLLL', header[ptr:ptr+52])
h_ofs_bmap   = part[0] # offset to blockmap, usually 0x100000
h_ofs_data   = part[1] # offset to blocks,   usually 0x200000
h_dc_cyls    = part[2] # unused
h_dc_heads   = part[3] # unused
h_dc_sectors = part[4] # unused
h_sz_sector  = part[5] # size of one sector, dontcare
h_x0         = part[6] # unused
h_sz_disk    = part[7] # size of entire physical disk in bytes
h_sz_block   = part[8] # size of one block in bytes
h_ex_block   = part[9] # block flags, dontcare
h_num_blocks = part[10] # total number of blocks in bmap
h_num_alloc  = part[11] # number of blocks in use which contain non-zero data
ptr += 52

h_id_image   = uuid(header[ptr+ 0:ptr+16])
h_id_snap    = uuid(header[ptr+16:ptr+32])
h_id_link    = uuid(header[ptr+32:ptr+48])
h_id_parent  = uuid(header[ptr+48:ptr+64])
ptr += 64

part = struct.unpack(b'QQQQQQQ', header[ptr:ptr+56])
h_x1         = part[0] #unused (  ff 0000 0400)
h_x2         = part[1] #unused ( 200 0000 003f)
h_x3         = part[2] #unused
h_x4         = part[3] #unused
h_x5         = part[4] #unused
h_x6         = part[5] #unused
h_x7         = part[6] #unused

if VERBOSE:
	dbg(h_ofs_bmap, h_ofs_data)
	dbg(h_dc_cyls, h_dc_heads, h_dc_sectors, h_sz_sector, h_x0)
	dbg(h_sz_disk, h_sz_block, h_ex_block, h_num_blocks, h_num_alloc)
	p('uuid_image   ' + h_id_image)
	p('uuid_snap    ' + h_id_snap)
	p('uuid_link    ' + h_id_link)
	p('uuid_parent  ' + h_id_parent)
	dbg(h_x1, h_x2, h_x3, h_x4, h_x5, h_x6, h_x7)

if h_id_link != '00000000-0000-0000-0000-000000000000':
	p('this vdi is a symlink to {0}'.format(h_id_link))
	p('nothing to do, exiting')
	exit()

if h_id_parent != '00000000-0000-0000-0000-000000000000':
	p('this vdi has a parent: {0}'.format(h_id_parent))
	p('nothing to do, exiting')
	exit()

sz_bmap = h_num_blocks * 4
nul = b'\0' * h_sz_block
remain = h_num_alloc
write_ptr = 0
bmap_ptr = 0
t0 = time.time()
tp = 0

p('loading map at [0x{0:x},0x{1:x}] ({2} entries)'.format(
	h_ofs_bmap, h_ofs_bmap + sz_bmap, h_num_blocks))

vdi.seek(h_ofs_bmap)
bmap = vdi.read(sz_bmap)

p('writing {0} data, {1} zeros, {2} total'.format(
	human(h_num_alloc*h_sz_block),
	human((h_num_blocks-h_num_alloc)*h_sz_block),
	human(h_num_blocks*h_sz_block)))

padding = ''
if h_num_blocks > 99999:
	padding = '  '

p('\n  disk%    {0}diskblock  (datablock)\n\n{1}\n\n'.format(
	padding, '.'*78))

blocks = [ u'\u258f', u'\u258e', u'\u258d', u'\u258c', u'\u258b', u'\u258a', u'\u2589', u'\u2588' ]
fullblock = blocks[len(blocks)-1]

while bmap_ptr < sz_bmap:
	t = time.time()
	if t > tp + 0.1 or remain < 0 or bmap_ptr >= sz_bmap - 4:
		tp = t
		perc = 100 * (bmap_ptr * 1.0 / sz_bmap)
		if False:
			p('\033[A{0:6.2f}%  {1:5}/{2}  ({3}/{4})\033[K'.format(
				perc, bmap_ptr/4, sz_bmap/4, remain, h_num_alloc))
		else:
			barwidth = perc * 0.78
			remainder = barwidth % 1
			bar = fullblock * int(barwidth)
			bar += blocks[int(remainder*len(blocks))]
			p(u'\033[4A{0:6.2f}%  {1:5}/{2:<5}  ({3}/{4})\033[K\n\n{5}\n'.format(
				perc, bmap_ptr/4, sz_bmap/4, remain, h_num_alloc, bar))

	bmap_entry = struct.unpack(b'<L', bmap[bmap_ptr:bmap_ptr+4])[0]
	bmap_ptr += 4
	# 0xffffffff is unallocated
	# 0xfffffffe is discarded (effectively unallocated)
	if bmap_entry >= 0xfffffffe:
		if is_stdout:
			stdout.write(nul)
		write_ptr += h_sz_block
		continue
	#p('{:x} = {:x}'.format(bmap_ptr, bmap_entry))
	
	seek_to = h_ofs_data + h_sz_block * bmap_entry
	if not vdi.tell() == seek_to:
		vdi.seek(seek_to)
	blk = vdi.read(h_sz_block)
	
	if not is_stdout:
		if not target.tell() == write_ptr:
			target.seek(write_ptr)
	target.write(blk)
	write_ptr += h_sz_block
	remain -= 1

t1 = time.time()
target.close()

p('\nfinished in {0:.3f} sec'.format(t1-t0))
p()

